"""
MIT License

Copyright (c) 2025 RenzMc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os


class RenzmcBuiltinFunction:

    def __init__(self, func, name):
        self.func = func
        self.name = name
        self.__name__ = name

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return f"<builtin function '{self.name}'>"


def baca_file(filename):
    try:
        with open(filename, "r", encoding="utf-8") as f:
            return f.read()
    except FileNotFoundError:
        raise FileNotFoundError(f"File '{filename}' tidak ditemukan")
    except Exception as e:
        raise Exception(f"Error membaca file: {e}")


def tulis_file(filename, content):
    try:
        with open(filename, "w", encoding="utf-8") as f:
            f.write(str(content))
        return True
    except Exception as e:
        raise Exception(f"Error menulis file: {e}")


def tambah_file(filename, content):
    try:
        with open(filename, "a", encoding="utf-8") as f:
            f.write(str(content))
        return True
    except Exception as e:
        raise Exception(f"Error menambah ke file: {e}")


def hapus_file(filename):
    try:
        os.remove(filename)
        return True
    except FileNotFoundError:
        raise FileNotFoundError(f"File '{filename}' tidak ditemukan")
    except Exception as e:
        raise Exception(f"Error menghapus file: {e}")


def gabung_path(*paths):
    return os.path.join(*paths)


def file_exists(path):
    return os.path.exists(path)


def buat_direktori(path):
    try:
        os.makedirs(path, exist_ok=True)
        return True
    except Exception as e:
        raise Exception(f"Error membuat direktori: {e}")


def daftar_direktori(path="."):
    try:
        return os.listdir(path)
    except FileNotFoundError:
        raise FileNotFoundError(f"Direktori '{path}' tidak ditemukan")
    except Exception as e:
        raise Exception(f"Error membaca direktori: {e}")


def direktori_ada_impl(path):
    return os.path.isdir(path)


def direktori_sekarang_impl():
    return os.getcwd()


def ubah_direktori_impl(path):
    try:
        os.chdir(path)
        return True
    except FileNotFoundError:
        raise FileNotFoundError(f"Direktori '{path}' tidak ditemukan")
    except Exception as e:
        raise Exception(f"Error mengubah direktori: {e}")


def pisah_path_impl(path):
    return os.path.split(path)


def ekstensi_file_impl(path):
    return os.path.splitext(path)[1]


def nama_file_tanpa_ekstensi_impl(path):
    return os.path.splitext(os.path.basename(path))[0]


def path_ada_impl(path):
    return os.path.exists(path)


def adalah_file_impl(path):
    return os.path.isfile(path)


def adalah_direktori_impl(path):
    return os.path.isdir(path)


def path_absolut_impl(path):
    return os.path.abspath(path)


def waktu_modifikasi_file_impl(path):
    try:
        return os.path.getmtime(path)
    except FileNotFoundError:
        raise FileNotFoundError(f"File '{path}' tidak ditemukan")


def waktu_buat_file_impl(path):
    try:
        return os.path.getctime(path)
    except FileNotFoundError:
        raise FileNotFoundError(f"File '{path}' tidak ditemukan")


def file_dapat_dibaca_impl(path):
    return os.access(path, os.R_OK)


def file_dapat_ditulis_impl(path):
    return os.access(path, os.W_OK)


def buka_impl(filename, mode="r", encoding="utf-8", **kwargs):
    try:
        return open(filename, mode, encoding=encoding, **kwargs)
    except FileNotFoundError:
        raise FileNotFoundError(f"File '{filename}' tidak ditemukan")
    except Exception as e:
        raise Exception(f"Error membuka file: {e}")


def tutup_impl(file_obj):
    try:
        file_obj.close()
        return True
    except Exception as e:
        raise Exception(f"Error menutup file: {e}")


def tulis_impl(file_obj, content):
    try:
        file_obj.write(str(content))
        return True
    except Exception as e:
        raise Exception(f"Error menulis ke file: {e}")


def baca_impl(file_obj, size=-1):
    try:
        return file_obj.read(size)
    except Exception as e:
        raise Exception(f"Error membaca file: {e}")


direktori_ada = RenzmcBuiltinFunction(direktori_ada_impl, "direktori_ada")
direktori_sekarang = RenzmcBuiltinFunction(direktori_sekarang_impl, "direktori_sekarang")
ubah_direktori = RenzmcBuiltinFunction(ubah_direktori_impl, "ubah_direktori")
pisah_path = RenzmcBuiltinFunction(pisah_path_impl, "pisah_path")
ekstensi_file = RenzmcBuiltinFunction(ekstensi_file_impl, "ekstensi_file")
nama_file_tanpa_ekstensi = RenzmcBuiltinFunction(nama_file_tanpa_ekstensi_impl, "nama_file_tanpa_ekstensi")
path_ada = RenzmcBuiltinFunction(path_ada_impl, "path_ada")
adalah_file = RenzmcBuiltinFunction(adalah_file_impl, "adalah_file")
adalah_direktori = RenzmcBuiltinFunction(adalah_direktori_impl, "adalah_direktori")
path_absolut = RenzmcBuiltinFunction(path_absolut_impl, "path_absolut")
waktu_modifikasi_file = RenzmcBuiltinFunction(waktu_modifikasi_file_impl, "waktu_modifikasi_file")
waktu_buat_file = RenzmcBuiltinFunction(waktu_buat_file_impl, "waktu_buat_file")
file_dapat_dibaca = RenzmcBuiltinFunction(file_dapat_dibaca_impl, "file_dapat_dibaca")
file_dapat_ditulis = RenzmcBuiltinFunction(file_dapat_ditulis_impl, "file_dapat_ditulis")
buka = RenzmcBuiltinFunction(buka_impl, "buka")
tutup = RenzmcBuiltinFunction(tutup_impl, "tutup")
tulis = RenzmcBuiltinFunction(tulis_impl, "tulis")
baca = RenzmcBuiltinFunction(baca_impl, "baca")



"""
MIT License

Copyright (c) 2025 RenzMc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class RenzmcBuiltinFunction:

    def __init__(self, func, name):
        self.func = func
        self.name = name
        self.__name__ = name

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return f"<builtin function '{self.name}'>"


def panjang(obj):
    try:
        return len(obj)
    except TypeError:
        raise TypeError(f"Objek tipe '{type(obj).__name__}' tidak memiliki panjang")


def jenis(obj):
    return type(obj).__name__


def ke_teks(obj):
    return str(obj)


def ke_angka(obj):
    try:
        return int(obj)
    except ValueError:
        try:
            return float(obj)
        except ValueError:
            raise ValueError(f"Tidak dapat mengkonversi '{obj}' ke angka")


def _convert_to_bool(value):
    if isinstance(value, bool):
        return value
    if isinstance(value, (int, float)):
        return value != 0
    if isinstance(value, str):
        return value.lower() in ("true", "1", "yes", "ya", "benar")
    return bool(value)


def input_impl(prompt=""):
    return input(prompt)


def print_impl(*args, sep=" ", end="\n"):
    print(*args, sep=sep, end=end)


def list_impl(iterable):
    import builtins as _builtins

    return _builtins.list(iterable)


def dict_impl(*args, **kwargs):
    import builtins as _builtins

    return _builtins.dict(*args, **kwargs)


def set_impl(iterable=None):
    import builtins as _builtins

    if iterable is None:
        return _builtins.set()
    return _builtins.set(iterable)


def tuple_impl(iterable=None):
    import builtins as _builtins

    if iterable is None:
        return ()
    return _builtins.tuple(iterable)


def str_impl(obj):
    import builtins as _builtins

    return _builtins.str(obj)


def int_impl(obj, base=10):
    import builtins as _builtins

    return _builtins.int(obj, base)


def float_impl(obj):
    import builtins as _builtins

    return _builtins.float(obj)


def bool_impl(obj):
    import builtins as _builtins

    return _builtins.bool(obj)


def sum_impl(iterable, start=0):
    import builtins as _builtins

    return _builtins.sum(iterable, start)


def len_impl(obj):
    import builtins as _builtins

    return _builtins.len(obj)


def min_impl(*args, **kwargs):
    import builtins as _builtins

    return _builtins.min(*args, **kwargs)


def max_impl(*args, **kwargs):
    import builtins as _builtins

    return _builtins.max(*args, **kwargs)


def abs_impl(x):
    import builtins as _builtins

    return _builtins.abs(x)


def round_impl(number, ndigits=None):
    import builtins as _builtins

    if ndigits is None:
        return _builtins.round(number)
    return _builtins.round(number, ndigits)


def pow_impl(base, exp, mod=None):
    import builtins as _builtins

    if mod is None:
        return _builtins.pow(base, exp)
    return _builtins.pow(base, exp, mod)


input_renzmc = RenzmcBuiltinFunction(input_impl, "input")
masukan = input_renzmc

print_renzmc = RenzmcBuiltinFunction(print_impl, "print")
cetak = print_renzmc

list_renzmc = RenzmcBuiltinFunction(list_impl, "list")
daftar = list_renzmc

dict_renzmc = RenzmcBuiltinFunction(dict_impl, "dict")
kamus = dict_renzmc

set_renzmc = RenzmcBuiltinFunction(set_impl, "set")
himpunan = set_renzmc

tuple_renzmc = RenzmcBuiltinFunction(tuple_impl, "tuple")
tupel = tuple_renzmc

str_renzmc = RenzmcBuiltinFunction(str_impl, "str")
teks = str_renzmc

int_renzmc = RenzmcBuiltinFunction(int_impl, "int")
bilangan_bulat = int_renzmc

float_renzmc = RenzmcBuiltinFunction(float_impl, "float")
bilangan_desimal = float_renzmc

bool_renzmc = RenzmcBuiltinFunction(bool_impl, "bool")
boolean = bool_renzmc

sum_renzmc = RenzmcBuiltinFunction(sum_impl, "sum")
jumlah_sum = sum_renzmc

len_renzmc = RenzmcBuiltinFunction(len_impl, "len")
panjang_len = len_renzmc

min_renzmc = RenzmcBuiltinFunction(min_impl, "min")
minimum_min = min_renzmc

max_renzmc = RenzmcBuiltinFunction(max_impl, "max")
maksimum_max = max_renzmc

abs_renzmc = RenzmcBuiltinFunction(abs_impl, "abs")
nilai_absolut = abs_renzmc

round_renzmc = RenzmcBuiltinFunction(round_impl, "round")
bulatkan = round_renzmc

pow_renzmc = RenzmcBuiltinFunction(pow_impl, "pow")
pangkat_pow = pow_renzmc



"""
MIT License

Copyright (c) 2025 RenzMc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""
RenzmcLang Interpreter Advanced Visitors Module

This module contains advanced visitors methods.
"""


from renzmc.core.ast import String

try:
    from renzmc.jit import JITCompiler

    JIT_AVAILABLE = True
except ImportError:
    JIT_AVAILABLE = False
    JITCompiler = None


class AdvancedVisitorsMixin:
    """
    Mixin class for advanced visitors.

    Provides 14 methods for handling advanced visitors.
    """

    def visit_IndexAccess(self, node):
        obj = self.visit(node.obj)
        index = self.visit(node.index)
        try:
            return obj[index]
        except (IndexError, KeyError):
            raise IndexError(
                f"Indeks '{index}' di luar jangkauan untuk objek tipe '{type(obj).__name__}'"
            )
        except TypeError:
            raise TypeError(f"Objek tipe '{type(obj).__name__}' tidak mendukung pengindeksan")

    def visit_SliceAccess(self, node):
        obj = self.visit(node.obj)
        start = self.visit(node.start) if node.start else None
        end = self.visit(node.end) if node.end else None
        step = self.visit(node.step) if node.step else None
        try:
            return obj[start:end:step]
        except TypeError:
            raise TypeError(f"Objek tipe '{type(obj).__name__}' tidak mendukung slicing")

    def visit_ListComp(self, node):
        var_name = node.var_name
        iterable = self.visit(node.iterable)
        if not hasattr(iterable, "__iter__"):
            raise TypeError(f"Objek tipe '{type(iterable).__name__}' tidak dapat diiterasi")
        result = []
        old_local_scope = self.local_scope.copy()
        for item in iterable:
            self.set_variable(var_name, item, is_local=True)
            if node.condition:
                condition_result = self.visit(node.condition)
                if not condition_result:
                    continue
            expr_result = self.visit(node.expr)
            result.append(expr_result)
        self.local_scope = old_local_scope
        return result

    def visit_DictComp(self, node):
        var_name = node.var_name
        iterable = self.visit(node.iterable)
        if not hasattr(iterable, "__iter__"):
            raise TypeError(f"Objek tipe '{type(iterable).__name__}' tidak dapat diiterasi")
        result = {}
        old_local_scope = self.local_scope.copy()
        for item in iterable:
            self.set_variable(var_name, item, is_local=True)
            if node.condition:
                condition_result = self.visit(node.condition)
                if not condition_result:
                    continue
            key_result = self.visit(node.key_expr)
            value_result = self.visit(node.value_expr)
            result[key_result] = value_result
        self.local_scope = old_local_scope
        return result

    def visit_SetComp(self, node):
        var_name = node.var_name
        iterable = self.visit(node.iterable)
        if not hasattr(iterable, "__iter__"):
            raise TypeError(f"Objek tipe '{type(iterable).__name__}' tidak dapat diiterasi")
        result = set()
        old_local_scope = self.local_scope.copy()
        for item in iterable:
            self.set_variable(var_name, item, is_local=True)
            if node.condition:
                condition_result = self.visit(node.condition)
                if not condition_result:
                    continue
            expr_result = self.visit(node.expr)
            result.add(expr_result)
        self.local_scope = old_local_scope
        return result

    def visit_Decorator(self, node):
        name = node.name
        args = [self.visit(arg) for arg in node.args]
        decorated = self.visit(node.decorated)
        if name in self.advanced_features.decorators:
            raw_decorator_func = self.advanced_features.decorators[name]
            try:
                from renzmc.runtime.advanced_features import RenzmcDecorator

                decorator_instance = RenzmcDecorator(raw_decorator_func, args)
                decorated_function = decorator_instance(decorated)

                # Check if this is a marker decorator
                marker_decorators = {"jit_compile", "jit_force", "gpu", "parallel"}

                if hasattr(node.decorated, "name"):
                    func_name = node.decorated.name

                    # For marker decorators, just set attributes on the function metadata
                    if name in marker_decorators:
                        # Store decorator hints in function metadata
                        if not hasattr(self, "_function_decorators"):
                            self._function_decorators = {}
                        if func_name not in self._function_decorators:
                            self._function_decorators[func_name] = []
                        self._function_decorators[func_name].append(name)

                        # Set attributes directly on the function if it exists
                        if func_name in self.functions:
                            # Mark the function with JIT hints
                            if name == "jit_compile":
                                if not hasattr(self, "_jit_hints"):
                                    self._jit_hints = set()
                                self._jit_hints.add(func_name)
                            elif name == "jit_force":
                                if not hasattr(self, "_jit_force"):
                                    self._jit_force = set()
                                self._jit_force.add(func_name)
                            elif name == "gpu":
                                if not hasattr(self, "_gpu_functions"):
                                    self._gpu_functions = set()
                                self._gpu_functions.add(func_name)
                            elif name == "parallel":
                                if not hasattr(self, "_parallel_functions"):
                                    self._parallel_functions = set()
                                self._parallel_functions.add(func_name)

                        # Don't add to _decorated_functions for marker decorators
                        return decorated_function

                    # For wrapper decorators (like @profile), store the wrapped function
                    self._decorated_functions = getattr(self, "_decorated_functions", {})

                    def original_func_callable(*call_args, **call_kwargs):
                        if func_name in self.functions:
                            params, body, return_type, param_types = self.functions[func_name]
                            return self._execute_user_function(
                                func_name,
                                params,
                                body,
                                return_type,
                                param_types,
                                call_args,
                                call_kwargs,
                            )
                        else:
                            raise NameError(f"Fungsi asli '{func_name}' tidak ditemukan")

                    original_func_callable.__name__ = func_name

                    # Apply the decorator to get the wrapped function
                    wrapped_function = raw_decorator_func(original_func_callable)

                    # Store the wrapped function directly
                    self._decorated_functions[func_name] = wrapped_function
                return decorated_function
            except Exception as e:
                raise RuntimeError(f"Error dalam dekorator '{name}': {str(e)}")
        if name in self.functions:
            try:
                if args:
                    decorator_factory = self._create_user_decorator_factory(name, args)
                    return decorator_factory(decorated)
                else:
                    user_decorator_func = self._create_user_function_wrapper(name)
                    from renzmc.runtime.advanced_features import RenzmcDecorator

                    decorator_instance = RenzmcDecorator(user_decorator_func)
                    return decorator_instance(decorated)
            except Exception as e:
                raise RuntimeError(f"Error dalam dekorator '{name}': {str(e)}")
        if hasattr(self, "decorators") and name in self.decorators:
            decorator_func = self.decorators[name]
            try:
                from renzmc.runtime.advanced_features import RenzmcDecorator

                decorator_instance = RenzmcDecorator(decorator_func, args)
                return decorator_instance(decorated)
            except Exception as e:
                raise RuntimeError(f"Error dalam dekorator '{name}': {str(e)}")
        raise NameError(f"Dekorator '{name}' tidak ditemukan")

    def visit_TypeHint(self, node):
        return node.type_name

    def visit_TypeAlias(self, node):
        self.type_registry[node.name] = node.type_expr
        return None

    def visit_LiteralType(self, node):
        return node

    def visit_TypedDictType(self, node):
        return node

    def visit_FormatString(self, node):
        result = ""
        for part in node.parts:
            if isinstance(part, String):
                result += part.value
            else:
                try:
                    value = self.visit(part)
                    if value is not None:
                        result += str(value)
                    else:
                        result += "None"
                except Exception as e:
                    result += f"<Error: {str(e)}>"
        return result

    def visit_Unpacking(self, node):
        value = self.visit(node.expr)
        if not hasattr(value, "__iter__"):
            raise TypeError(f"Objek tipe '{type(value).__name__}' tidak dapat diiterasi")
        return value

    def visit_ExtendedUnpacking(self, node):
        value = self.visit(node.value)
        if not isinstance(value, (list, tuple)):
            try:
                value = list(value)
            except (TypeError, ValueError) as e:
                self.error(
                    f"Nilai tidak dapat di-unpack: {type(value).__name__} - {e}",
                    node.token,
                )
        starred_index = None
        for i, (name, is_starred) in enumerate(node.targets):
            if is_starred:
                if starred_index is not None:
                    self.error(
                        "Hanya satu target yang dapat menggunakan * dalam unpacking",
                        node.token,
                    )
                starred_index = i
        num_targets = len(node.targets)
        num_values = len(value)
        if starred_index is None:
            if num_targets != num_values:
                self.error(
                    f"Jumlah nilai ({num_values}) tidak sesuai dengan jumlah target ({num_targets})",
                    node.token,
                )
            for (name, _), val in zip(node.targets, value):
                self.current_scope.set(name, val)
        else:
            num_required = num_targets - 1
            if num_values < num_required:
                self.error(
                    f"Tidak cukup nilai untuk unpack (dibutuhkan minimal {num_required}, ada {num_values})",
                    node.token,
                )
            for i in range(starred_index):
                name, _ = node.targets[i]
                self.current_scope.set(name, value[i])
            num_after_starred = num_targets - starred_index - 1
            starred_count = num_values - num_required
            starred_name, _ = node.targets[starred_index]
            starred_values = value[starred_index : starred_index + starred_count]
            self.current_scope.set(starred_name, list(starred_values))
            for i in range(num_after_starred):
                target_index = starred_index + 1 + i
                value_index = starred_index + starred_count + i
                name, _ = node.targets[target_index]
                self.current_scope.set(name, value[value_index])

    def visit_StarredExpr(self, node):
        value = self.visit(node.expr)
        if isinstance(value, (list, tuple)):
            return value
        try:
            return list(value)
        except (TypeError, ValueError) as e:
            self.error(f"Nilai tidak dapat di-unpack: {type(value).__name__} - {e}", node.token)



"""
MIT License

Copyright (c) 2025 RenzMc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
import base64
import datetime
import hashlib
import importlib
import inspect
import json
import os
import re
import sys
import time
import urllib.parse
import urllib.request
import uuid
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union


def is_identifier(name):
    return re.match("^[a-zA-Z_][a-zA-Z0-9_]*$", name) is not None


def is_keyword(name):
    keywords = [
        "jika",
        "kalau",
        "maka",
        "tidak",
        "lainnya",
        "selesai",
        "selama",
        "ulangi",
        "kali",
        "untuk",
        "setiap",
        "dari",
        "sampai",
        "lanjut",
        "berhenti",
        "coba",
        "tangkap",
        "akhirnya",
        "simpan",
        "ke",
        "dalam",
        "itu",
        "adalah",
        "sebagai",
        "tampilkan",
        "tulis",
        "cetak",
        "tunjukkan",
        "tanya",
        "buat",
        "fungsi",
        "dengan",
        "parameter",
        "panggil",
        "jalankan",
        "kembali",
        "hasil",
        "kelas",
        "metode",
        "konstruktor",
        "warisi",
        "gunakan",
        "impor",
        "impor_python",
        "modul",
        "paket",
        "lambda",
        "async",
        "await",
        "yield",
        "dekorator",
        "tipe",
        "jenis_data",
        "generator",
        "asinkron",
        "dan",
        "atau",
        "benar",
        "salah",
        "self",
        "ini",
    ]
    return name in keywords


def format_code(code):
    lines = code.split("\n")
    result = []
    indent_level = 0
    for line in lines:
        stripped = line.strip()
        if stripped in [
            "selesai",
            "kalau tidak",
            "kalau tidak jika",
            "akhirnya",
            "tangkap",
        ]:
            indent_level = max(0, indent_level - 1)
        if stripped:
            result.append("    " * indent_level + stripped)
        else:
            result.append("")
        if stripped.endswith(":") or any(
            (
                stripped.startswith(keyword)
                for keyword in [
                    "jika",
                    "kalau tidak jika",
                    "kalau tidak",
                    "selama",
                    "untuk setiap",
                    "ulangi",
                    "buat fungsi",
                    "buat kelas",
                    "buat metode",
                    "konstruktor",
                    "coba",
                    "tangkap",
                ]
            )
        ):
            indent_level += 1
    return "\n".join(result)


def parse_type_annotation(annotation):
    if annotation == "int" or annotation == "bilangan_bulat":
        return int
    elif annotation == "float" or annotation == "desimal":
        return float
    elif annotation == "str" or annotation == "teks":
        return str
    elif annotation == "bool" or annotation == "boolean":
        return bool
    elif annotation == "list" or annotation == "daftar":
        return list
    elif annotation == "dict" or annotation == "kamus":
        return dict
    elif annotation == "set" or annotation == "himpunan":
        return set
    elif annotation == "tuple" or annotation == "tupel":
        return tuple
    elif annotation == "None" or annotation == "kosong":
        return type(None)
    elif annotation == "Any" or annotation == "apapun":
        return Any
    elif annotation.startswith("List[") or annotation.startswith("Daftar["):
        inner = annotation[annotation.index("[") + 1 : annotation.rindex("]")]
        return List[parse_type_annotation(inner)]
    elif annotation.startswith("Dict[") or annotation.startswith("Kamus["):
        inner = annotation[annotation.index("[") + 1 : annotation.rindex("]")]
        key_type, value_type = inner.split(",")
        return Dict[
            parse_type_annotation(key_type.strip()),
            parse_type_annotation(value_type.strip()),
        ]
    elif annotation.startswith("Set[") or annotation.startswith("Himpunan["):
        inner = annotation[annotation.index("[") + 1 : annotation.rindex("]")]
        return Set[parse_type_annotation(inner)]
    elif annotation.startswith("Tuple[") or annotation.startswith("Tupel["):
        inner = annotation[annotation.index("[") + 1 : annotation.rindex("]")]
        types = [parse_type_annotation(t.strip()) for t in inner.split(",")]
        return Tuple[tuple(types)]
    elif annotation.startswith("Optional[") or annotation.startswith("Opsional["):
        inner = annotation[annotation.index("[") + 1 : annotation.rindex("]")]
        return Optional[parse_type_annotation(inner)]
    elif annotation.startswith("Union[") or annotation.startswith("Gabungan["):
        inner = annotation[annotation.index("[") + 1 : annotation.rindex("]")]
        types = [parse_type_annotation(t.strip()) for t in inner.split(",")]
        return Union[tuple(types)]
    elif annotation.startswith("Callable[") or annotation.startswith("Fungsi["):
        return Callable
    else:
        return annotation


def check_type(value, type_annotation):
    if type_annotation is Any:
        return True
    if isinstance(type_annotation, str):
        type_annotation = parse_type_annotation(type_annotation)
    if isinstance(type_annotation, type):
        return isinstance(value, type_annotation)
    origin = getattr(type_annotation, "__origin__", None)
    args = getattr(type_annotation, "__args__", None)
    if origin is list or origin is List:
        return isinstance(value, list) and all((check_type(item, args[0]) for item in value))
    elif origin is dict or origin is Dict:
        return isinstance(value, dict) and all(
            (check_type(k, args[0]) and check_type(v, args[1]) for k, v in value.items())
        )
    elif origin is set or origin is Set:
        return isinstance(value, set) and all((check_type(item, args[0]) for item in value))
    elif origin is tuple or origin is Tuple:
        return (
            isinstance(value, tuple)
            and len(value) == len(args)
            and all((check_type(value[i], args[i]) for i in range(len(args))))
        )
    elif origin is Union:
        return any((check_type(value, arg) for arg in args))
    elif origin is Optional:
        return value is None or check_type(value, args[0])
    elif origin is Callable:
        return callable(value)
    return False


def format_error_message(error, source_code=None):
    if not hasattr(error, "line") or not hasattr(error, "column") or error.line is None or (error.column is None):
        return str(error)
    result = f"Error: {error.message}\n"
    result += f"Pada baris {error.line}, kolom {error.column}\n"
    code_to_use = error.source_code if hasattr(error, "source_code") and error.source_code else source_code
    if code_to_use:
        lines = code_to_use.split("\n")
        if 0 <= error.line - 1 < len(lines):
            line = lines[error.line - 1]
            result += f"\n{error.line} | {line}\n"
            result += " " * (len(str(error.line)) + 3 + error.column - 1) + "^\n"
            context_lines = 2
            start_line = max(0, error.line - 1 - context_lines)
            end_line = min(len(lines), error.line - 1 + context_lines + 1)
            if start_line > 0:
                result += "...\n"
            for i in range(start_line, end_line):
                if i == error.line - 1:
                    continue
                result += f"{i + 1} | {lines[i]}\n"
            if end_line < len(lines):
                result += "...\n"
    return result


def load_module(module_name):
    try:
        return importlib.import_module(module_name)
    except ImportError:
        raise ImportError(f"Tidak dapat mengimpor modul '{module_name}'")


def get_module_functions(module):
    return {name: obj for name, obj in inspect.getmembers(module, inspect.isfunction) if not name.startswith("_")}


def get_module_classes(module):
    return {name: obj for name, obj in inspect.getmembers(module, inspect.isclass) if not name.startswith("_")}


def get_module_variables(module):
    return {
        name: obj
        for name, obj in inspect.getmembers(module)
        if not name.startswith("_") and (not inspect.isfunction(obj)) and (not inspect.isclass(obj))
    }


def get_class_methods(cls):
    return {name: obj for name, obj in inspect.getmembers(cls, inspect.isfunction) if not name.startswith("_")}


def get_class_attributes(cls):
    return {
        name: obj
        for name, obj in inspect.getmembers(cls)
        if not name.startswith("_") and (not inspect.isfunction(obj)) and (not inspect.ismethod(obj))
    }


def get_function_signature(func):
    return str(inspect.signature(func))


def get_function_parameters(func):
    return list(inspect.signature(func).parameters.keys())


def get_function_defaults(func):
    signature = inspect.signature(func)
    return {
        name: param.default
        for name, param in signature.parameters.items()
        if param.default is not inspect.Parameter.empty
    }


def get_function_annotations(func):
    return func.__annotations__


def get_function_doc(func):
    return func.__doc__


def get_function_source(func):
    return inspect.getsource(func)


def is_async_function(func):
    return asyncio.iscoroutinefunction(func)


def run_async(coro):
    return asyncio.run(coro)


def wait_all_async(coros):
    return asyncio.run(asyncio.gather(*coros))


def create_async_function(func):

    async def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def json_to_dict(json_str):
    try:
        return json.loads(json_str)
    except json.JSONDecodeError as e:
        raise ValueError(f"JSON tidak valid: {str(e)}")


def dict_to_json(dictionary):
    try:
        return json.dumps(dictionary, ensure_ascii=False)
    except TypeError as e:
        raise TypeError(f"Tidak dapat mengkonversi kamus ke JSON: {str(e)}")


def get_current_time():
    return datetime.datetime.now()


def format_time(dt, format_str="%Y-%m-%d %H:%M:%S"):
    return dt.strftime(format_str)


def parse_time(time_str, format_str="%Y-%m-%d %H:%M:%S"):
    try:
        return datetime.datetime.strptime(time_str, format_str)
    except ValueError:
        raise ValueError(f"Format waktu tidak valid: '{time_str}' (format: '{format_str}')")


def get_timestamp():
    return time.time()


def timestamp_to_datetime(timestamp):
    return datetime.datetime.fromtimestamp(timestamp)


def datetime_to_timestamp(dt):
    return dt.timestamp()


def hash_string(string, algorithm="sha256"):
    if algorithm == "md5":
        return hashlib.md5(string.encode()).hexdigest()
    elif algorithm == "sha1":
        return hashlib.sha1(string.encode()).hexdigest()
    elif algorithm == "sha256":
        return hashlib.sha256(string.encode()).hexdigest()
    elif algorithm == "sha512":
        return hashlib.sha512(string.encode()).hexdigest()
    else:
        raise ValueError(f"Algoritma hash tidak valid: '{algorithm}'")


def generate_uuid():
    return str(uuid.uuid4())


def base64_encode(string):
    return base64.b64encode(string.encode()).decode()


def base64_decode(string):
    try:
        return base64.b64decode(string.encode()).decode()
    except Exception:
        raise ValueError(f"String Base64 tidak valid: '{string}'")


def url_encode(string):
    return urllib.parse.quote(string)


def url_decode(string):
    return urllib.parse.unquote(string)


def http_get(url, headers=None):
    req = urllib.request.Request(url, headers=headers or {})
    with urllib.request.urlopen(req) as response:
        return {
            "status": response.status,
            "headers": dict(response.headers),
            "content": response.read().decode("utf-8"),
        }


def http_post(url, data, headers=None):
    data_bytes = urllib.parse.urlencode(data).encode("utf-8")
    req = urllib.request.Request(url, data=data_bytes, headers=headers or {}, method="POST")
    with urllib.request.urlopen(req) as response:
        return {
            "status": response.status,
            "headers": dict(response.headers),
            "content": response.read().decode("utf-8"),
        }


def file_exists(path):
    return os.path.exists(path)


def is_file(path):
    return os.path.isfile(path)


def is_directory(path):
    return os.path.isdir(path)


def get_file_size(path):
    if not os.path.exists(path):
        raise FileNotFoundError(f"File tidak ditemukan: '{path}'")
    return os.path.getsize(path)


def get_file_modification_time(path):
    if not os.path.exists(path):
        raise FileNotFoundError(f"File tidak ditemukan: '{path}'")
    return os.path.getmtime(path)


def list_directory(path="."):
    if not os.path.exists(path):
        raise FileNotFoundError(f"Direktori tidak ditemukan: '{path}'")
    if not os.path.isdir(path):
        raise NotADirectoryError(f"Bukan direktori: '{path}'")
    return os.listdir(path)


def create_directory(path):
    if os.path.exists(path):
        if os.path.isdir(path):
            raise FileExistsError(f"Direktori sudah ada: '{path}'")
        else:
            raise FileExistsError(f"File sudah ada dengan nama yang sama: '{path}'")
    os.makedirs(path)


def remove_file(path):
    if not os.path.exists(path):
        raise FileNotFoundError(f"File tidak ditemukan: '{path}'")
    if os.path.isdir(path):
        raise IsADirectoryError(f"Tidak dapat menghapus direktori dengan fungsi ini: '{path}'")
    os.remove(path)


def remove_directory(path):
    if not os.path.exists(path):
        raise FileNotFoundError(f"Direktori tidak ditemukan: '{path}'")
    if not os.path.isdir(path):
        raise NotADirectoryError(f"Bukan direktori: '{path}'")
    os.rmdir(path)


def join_path(*paths):
    return os.path.join(*paths)


def get_absolute_path(path):
    return os.path.abspath(path)


def get_basename(path):
    return os.path.basename(path)


def get_dirname(path):
    return os.path.dirname(path)


def get_extension(path):
    return os.path.splitext(path)[1]


def change_extension(path, extension):
    return os.path.splitext(path)[0] + extension


def normalize_path(path):
    return os.path.normpath(path)


def get_current_directory():
    return os.getcwd()


def change_directory(path):
    if not os.path.exists(path):
        raise FileNotFoundError(f"Direktori tidak ditemukan: '{path}'")
    if not os.path.isdir(path):
        raise NotADirectoryError(f"Bukan direktori: '{path}'")
    os.chdir(path)


def get_environment_variable(name):
    if name not in os.environ:
        raise KeyError(f"Variabel lingkungan tidak ditemukan: '{name}'")
    return os.environ[name]


def set_environment_variable(name, value):
    os.environ[name] = value


def get_python_version():
    return sys.version


def get_platform():
    return sys.platform


def get_executable():
    return sys.executable


def get_path():
    return sys.path


def add_to_path(path):
    sys.path.append(path)


def get_modules():
    return sys.modules


def get_arguments():
    return sys.argv


def exit(code=0):
    sys.exit(code)



"""
MIT License

Copyright (c) 2025 RenzMc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json


class RenzmcBuiltinFunction:

    def __init__(self, func, name):
        self.func = func
        self.name = name
        self.__name__ = name

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return f"<builtin function '{self.name}'>"


def json_ke_teks(obj, indent=None):
    try:
        return json.dumps(obj, indent=indent, ensure_ascii=False)
    except Exception as e:
        raise ValueError(f"Error mengkonversi ke JSON: {e}")


def teks_ke_json(text):
    try:
        return json.loads(text)
    except json.JSONDecodeError as e:
        raise ValueError(f"Error parsing JSON: {e}")


def tulis_json_impl(filename, data, indent=2):
    try:
        with open(filename, "w", encoding="utf-8") as f:
            json.dump(data, f, indent=indent, ensure_ascii=False)
        return True
    except Exception as e:
        raise Exception(f"Error menulis JSON ke file: {e}")


def baca_json_impl(filename):
    try:
        with open(filename, "r", encoding="utf-8") as f:
            return json.load(f)
    except FileNotFoundError:
        raise FileNotFoundError(f"File '{filename}' tidak ditemukan")
    except json.JSONDecodeError as e:
        raise ValueError(f"Error parsing JSON dari file: {e}")


def ke_json_impl(data, indent=None):
    try:
        return json.dumps(data, indent=indent, ensure_ascii=False)
    except Exception as e:
        raise ValueError(f"Error mengkonversi ke JSON: {e}")


def dari_json_impl(json_string):
    try:
        return json.loads(json_string)
    except json.JSONDecodeError as e:
        raise ValueError(f"Error parsing JSON: {e}")


tulis_json = RenzmcBuiltinFunction(tulis_json_impl, "tulis_json")
baca_json = RenzmcBuiltinFunction(baca_json_impl, "baca_json")
ke_json = RenzmcBuiltinFunction(ke_json_impl, "ke_json")
dari_json = RenzmcBuiltinFunction(dari_json_impl, "dari_json")
